/*!
 * \brief Argument parser for the demo application
 * \copyright @ 2022 RnD Center ELVEES, JSC
 */

#include <iomanip>
#include <iostream>
#include <string>
#include <vector>
#include <memory>
#include <regex>

#include <arg_parser.h>

ArgsParser* ArgsParser::Parse(int argc, char** argv) {
  auto* parser = new ArgsParser;
  std::string framerate = "";
  int input_argv_pos = 0;
  for (int i = 1; i < argc; i += 2) {
    std::string key(argv[i]);
    if (key == "-h" || key == "--help") {
      auto msg = ArgsParser::GetHelpMessage();
      std::cout << msg << std::endl;
      delete parser;
      return nullptr;
    } else if (key == "-i") {
      // save position of argv
      input_argv_pos = i + 1;
    } else if (key == "-o") {
      if (!parser->ParseOutput(argv[i + 1])) {
        break;
      }
    } else if (key == "-d") {
      parser->examples_dir = argv[i + 1];
    } else if (key == "--dsp-id") {
      parser->dsp_id = atoi(argv[i + 1]);
    } else if (key == "--framerate") {
      framerate = argv[i + 1];
    } else {
      auto msg =
          "Unrecognized program option: " + key + ArgsParser::GetHelpMessage();
      std::cout << msg << std::endl;
      delete parser;
      return nullptr;
    }
  }

  parser->ParseInput(argv[input_argv_pos], framerate);

  if (!parser->input_cfg || !parser->output_cfg) {
    auto msg = "Can't parse input parameters" + ArgsParser::GetHelpMessage();
    std::cout << msg << std::endl;
    delete parser;
    return nullptr;
  }

  return parser;
}

std::string ArgsParser::GetInputPipelineString() const {
  return input_cfg->Get() + " ! appsink";
}

std::string ArgsParser::GetOutputPipelineString() const {
  return "appsrc ! " + output_cfg->Get();
}

uint32_t ArgsParser::GetDSPId() const {
  return dsp_id;
}

const std::string& ArgsParser::GetExamplesDir() const {
  return examples_dir;
}

const std::string ArgsParser::GetHelpMessage() {
  unsigned int alignment_args = 15;
  std::stringstream help_message;
  help_message << "\n\nObject classification from video\n";
  help_message
      << "Usage: demo_classification -i [VIDEO_PATH] -o [OUTPUT_METHOD]\n\n";
  help_message << "Arguments:\n";

  help_message.setf(std::ios::left, std::ios::adjustfield);
  help_message.width(alignment_args);
  help_message << "VIDEO_PATH";
  help_message << " - path to source video in h264 format";
  help_message << std::endl;

  help_message.width(alignment_args);
  help_message << "OUTPUT_METHOD";
  help_message << " - video output method. Supported methods:";
  help_message << std::endl;

  help_message << " hdmi@mcom03-cfg - output to MCom03 HDMI port";
  help_message << std::endl;

  help_message << " rtsp://<ip>:><port>/<path> - output to RTSP server";
  help_message << std::endl;

  help_message << " <filename>.h264 - path to result video in h264 format";
  help_message << std::endl;

  help_message << "-d path to elcorenn-examples-data. Default: "
                  "$HOME/elcorenn-examples-data";
  help_message << std::endl;

  help_message << "--dsp-id select dsp to calculate test. Ex: --dsp-id 0";
  help_message << std::endl;

  help_message
      << "--framerate changes framerate of input stream. Ex: --framerate 24/1";
  help_message << std::endl;

  return help_message.str();
}

bool ArgsParser::ParseInput(std::string str, std::string framerate) {
  std::string video_format = ".h264";
  if (str.rfind(video_format) ==
      std::abs(static_cast<int>(str.size() - video_format.size()))) {
    input_cfg = std::make_shared<InputConfig>(str, framerate);
    return true;
  }

  return false;
}

bool ArgsParser::ParseOutput(std::string str) {
  std::string video_format = ".h264";
  if (str == "hdmi@mcom03-cfg") {
    output_cfg = std::make_shared<OutputConfigHdmiMCom03>();
    return true;
  } else if (str.rfind("rtsp://") == 0) {
    output_cfg = std::make_shared<OutputConfigRtsp>(str);
    return true;
  } else if (str.rfind(video_format) ==
             std::abs(static_cast<int>(str.size() - video_format.size()))) {
    output_cfg = std::make_shared<OutputConfigFile>(str);
    return true;
  } else if (str == "hdmi@solaris-cfg") {
    // HDMI isn't support as a Solaris output device #SOLARISSW-1771
    return false;
  }

  return false;
}
