/*!
 * \brief Demo application of classification frames from video
 * \copyright @ 2022 RnD Center ELVEES, JSC
 */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <memory>

#include <future>
#include <mutex>

#include <opencv2/core.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/videoio.hpp>

#include <elcorenn/elcorenn.h>

#include "arg_parser.h"
#include "imagenet_dataclasses.h"

using namespace cv;
using namespace std;

// threads-shared variables
Mutex mutex_results, mutex_frame;
double shared_predicted_probability = 0;
uint32_t shared_inference_time_ms = 0;
int shared_predicted_class_id = 0;
Mat shared_frame_original;

uint32_t dsp_id = 0;

// ImageNet preprocess function for MobileNet
void PreprocessFrameMobileNet(const Mat& input, Mat& output) {
  cv::cvtColor(input, output, cv::COLOR_BGR2RGB);
  // Preprocess frame
  resize(output, output, Size(224, 224), 0, 0);
  output.convertTo(output, CV_32FC3);
  output /= Mat3f(224, 224, Vec3f(127.5f, 127.5f, 127.5f));
  output -= Mat3f(224, 224, Vec3f(1.0f, 1.0f, 1.0f));
}

void SetSharedFrameOriginal(Mat& frame_original) {
  mutex_frame.lock();
  shared_frame_original = frame_original.clone();
  mutex_frame.unlock();
}

void GetSharedFrameOriginal(Mat& frame_original) {
  mutex_frame.lock();
  frame_original = shared_frame_original.clone();
  mutex_frame.unlock();
}

void SetSharedResults(int& predicted_class_id, double& predicted_probability,
                      uint32_t& inference_time_ms) {
  mutex_results.lock();
  shared_predicted_class_id = predicted_class_id;
  shared_predicted_probability = predicted_probability;
  shared_inference_time_ms = inference_time_ms;
  mutex_results.unlock();
}

void GetSharedResults(int& predicted_class_id, double& predicted_probability,
                      uint32_t& inference_time_ms) {
  mutex_results.lock();
  predicted_class_id = shared_predicted_class_id;
  predicted_probability = shared_predicted_probability;
  inference_time_ms = shared_inference_time_ms;
  mutex_results.unlock();
}

// Return average inference time based on last measurements
uint32_t GetAverageTime(uint32_t current_time) {
  static vector<uint32_t> time;
  static uint32_t index = 0;
  int max_size = 30;
  if (time.size() < max_size) {
    time.push_back(current_time);
  } else {
    time[index++] = current_time;
    index = index % max_size;
  }

  uint32_t time_sum = 0;
  for (auto& t : time) {
    time_sum += t;
  }
  return time_sum / time.size();
}

// draw-thread function
void DrawFrameProcess(VideoCapture& capture, VideoWriter& writer) {
  Mat frame_original, frame_resized;
  double predicted_probability = 0;
  uint32_t inference_time_ms = 0;
  uint32_t average_inference_time_ms = 0;
  int predicted_class_id = 0;
  vector<double> fps_vector;

  while (true) {
    capture >> frame_original;
    if (frame_original.empty()) {
      capture.set(CAP_PROP_POS_FRAMES, 0);
      continue;
    }
    // Convert frame format from NV12 to BGR
    cv::cvtColor(frame_original, frame_original, cv::COLOR_YUV2BGR_NV12);

    // Copy current frame to threads-shared variable
    SetSharedFrameOriginal(frame_original);

    Size frame_size = frame_original.size();

    // Draw rectangle
    rectangle(frame_original, Point(0, frame_size.height - 50),
              Point(frame_size.width, frame_size.height), Scalar(60, 60, 60),
              -1);

    // Get current results
    GetSharedResults(predicted_class_id, predicted_probability,
                     inference_time_ms);

    // Put results on original frame
    stringstream probability_str;
    probability_str << setprecision(2) << fixed << predicted_probability;

    const string& class_name = imagenet_class.at(predicted_class_id);
    string result_str = class_name.substr(0, class_name.find(',')) + " " +
                        probability_str.str();

    putText(frame_original, result_str, Point(20, frame_size.height - 15),
            FONT_HERSHEY_SIMPLEX, 0.5, Scalar(255, 255, 255), 1);

    if (inference_time_ms != 0) {
      average_inference_time_ms = GetAverageTime(inference_time_ms);
    }

    putText(frame_original,
            "Inference time, dsp" + to_string(dsp_id) + ": " +
                to_string(average_inference_time_ms) + " ms.",
            Point(frame_size.width - 250, frame_size.height - 15),
            FONT_HERSHEY_SIMPLEX, 0.5, Scalar(255, 255, 255), 1);

    writer << frame_original;
  }
}

int main(int argc, char** argv) {
  auto parser = ArgsParser::Parse(argc, argv);
  if (!parser) {
    return -1;
  }

  auto input_pipeline = parser->GetInputPipelineString();
  auto output_pipeline = parser->GetOutputPipelineString();
  dsp_id = parser->GetDSPId();
  uint32_t dsp_list[1] = {dsp_id};

  // Init ElcoreNN backend
  auto backend_id = InitBackendWithDevicesAssignment(1, dsp_list);

  auto examples_dir = parser->GetExamplesDir();
  std::string model_json = examples_dir + "/models/mobilenet_model.json";
  std::string model_weights = examples_dir + "/models/mobilenet_weights.bin";

  // Load model to DSP core (ELcore50)
  auto model_id = LoadModel(model_json.c_str(), model_weights.c_str(),
                            ENNDataType::FLOAT16, backend_id);

  // Create VideoCapture object using Gstreamer pipeline
  auto capture = VideoCapture(input_pipeline);
  if (!capture.isOpened()) {
    throw std::ifstream::failure(
        "Could not open video stream with GStreamer pipeline");
  }

  // Get frame rate, width and height from input video stream
  auto frame_width = static_cast<int>(capture.get(CAP_PROP_FRAME_WIDTH));
  auto frame_height = static_cast<int>(capture.get(CAP_PROP_FRAME_HEIGHT));
  auto frame_fps = capture.get(CAP_PROP_FPS);

  // Create VideoWriter object using Gstreamer pipeline
  auto writer = VideoWriter(output_pipeline, 0, frame_fps,
                            Size(frame_width, frame_height), true);

  if (!writer.isOpened()) {
    throw std::ifstream::failure(
        "Could not open VideoWriter with GStreamer pipeline");
  }

  // Create child process to draw results
  auto draw_thread =
      async(launch::async, DrawFrameProcess, ref(capture), ref(writer));

  while (draw_thread.wait_for(0ms) != future_status::ready) {
    Mat frame;

    // Get current frame from thread-shared variable
    GetSharedFrameOriginal(frame);

    // Check if draw-thread put data into global_frame_original
    if (!frame.empty()) {
      PreprocessFrameMobileNet(frame, frame);

      // Network output is 1000-element vector of probabilities
      Mat dst = Mat::zeros(1000, 1, CV_32FC1);

      float* inputs[1] = {frame.ptr<float>(0)};
      float* outputs[1] = {dst.ptr<float>(0)};
      // Measure inference time
      auto begin = chrono::steady_clock::now();
      // Run inference
      InvokeModel(model_id, inputs, outputs, 1);
      auto end = chrono::steady_clock::now();

      auto inference_time_ms = static_cast<uint32_t>(
          chrono::duration_cast<chrono::milliseconds>(end - begin).count());

      double minVal, maxVal;
      int minLoc, maxLoc;

      // Get max probability and find corresponding class
      minMaxIdx(dst, &minVal, &maxVal, &minLoc, &maxLoc);

      // Set results to threads-shared variables
      SetSharedResults(maxLoc, maxVal, inference_time_ms);
    }
  }

  // Release resources
  capture.release();
  writer.release();
  delete parser;
  ReleaseModel(model_id);
  ReleaseBackend(backend_id);

  return 0;
}
