// Inference example of YOLOv5s

#include <iostream>
#include <cmath>
#include <string>
#include <chrono>

#include "output_proc.hpp"

double score_threshold_default = 0.55;
double nms_threshold_default = 0.6;
double confidence_threshold_default = 0.4;

cv::Mat preprocess_image(cv::Mat& image, cv::Size& origin_size) {
  int fill_size = std::max(image.size().width, image.size().height) -
                  std::min(image.size().width, image.size().height);

  cv::Mat preprocessed_img;

  // Preprocess opened image
  cv::Scalar color = cv::Scalar(0, 0, 0); // black color
  if (image.size().width < image.size().height) {
    // add black border on the right
    cv::copyMakeBorder(image, preprocessed_img, 0, 0, 0, fill_size,
                       cv::BORDER_CONSTANT, color);
  } else {
    // add black border on the bottom
    cv::copyMakeBorder(image, preprocessed_img, 0, fill_size, 0, 0,
                       cv::BORDER_CONSTANT, color);
  }

  resize(preprocessed_img, preprocessed_img, cv::Size(640, 640), 0, 0);

  // change BGR format to RGB
  std::vector<cv::Mat> channels;
  cv::split(preprocessed_img, channels);
  std::swap(channels[2], channels[0]);

  cv::vconcat(channels, preprocessed_img);
  preprocessed_img = preprocessed_img.reshape(1, 1);

  preprocessed_img.convertTo(preprocessed_img, CV_32FC3);
  preprocessed_img /= 255;

  return preprocessed_img;
}

void print_usage(const std::string& name) {
  std::cout << "Usage: " << name << "[OPTIONS]" << std::endl
            << "Options description:" << std::endl
            << "    -i   Path to input image (JPG, etc.)" << std::endl
            << "    -o   Path to output file. "
            << "Default: pred_image.jpg" << std::endl
            << "    -d   Path to elcorenn-examples-data. "
            << "Default: $HOME/elcorenn-examples-data" << std::endl
            << "    -st  Score threshold value. "
            << "Default: " << score_threshold_default << std::endl
            << "    -iou NMS IOU threshold. "
            << "Default: " << nms_threshold_default << std::endl
            << "    -ct  Confidence threshold. "
            << "Default: " << confidence_threshold_default << std::endl
            << "    -h   Print usage" << std::endl;
}

int main(int argc, char** argv) {
  double score_threshold = score_threshold_default;
  double nms_threshold = nms_threshold_default;
  double confidence_threshold = confidence_threshold_default;

  std::string home_dir = getenv("HOME");
  std::string examples_dir = home_dir + "/elcorenn-examples-data";

  std::string output_file_path = "";
  std::string input_file_path = "";

  for (int i = 1; i < argc; i += 2) {
    std::string key(argv[i]);
    if (key == "-h") {
      print_usage(argv[0]);
      return 0;
    } else if (key == "-st")
      try {
        score_threshold = std::stod((std::string(argv[i + 1])));
      } catch (...) {
        throw std::runtime_error("Error: expected threshold value");
      }
    else if (key == "-iou")
      try {
        nms_threshold = std::stod((std::string(argv[i + 1])));
      } catch (...) {
        throw std::runtime_error("Error: expected threshold value");
      }
    else if (key == "-ct")
      try {
        confidence_threshold = std::stod((std::string(argv[i + 1])));
      } catch (...) {
        throw std::runtime_error("Error: expected threshold value");
      }
    else if (key == "-o") {
      try {
        output_file_path = std::string(argv[i + 1]);
      } catch (...) {
        throw std::runtime_error("Error: expected output file path");
      }
    } else if (key == "-i") {
      try {
        input_file_path = std::string(argv[i + 1]);
      } catch (...) {
        throw std::runtime_error("Error: expected input file path");
      }
    } else if (key == "-d") {
      try {
        examples_dir = std::string(argv[i + 1]);
      } catch (...) {
        throw std::runtime_error("Error: Set path to elcorenn-examples-data");
      }
    } else {
      std::cout << "Unrecognized program option: " << key << std::endl;
      print_usage(argv[0]);
      return 1;
    }
  }

  std::string output_file_path_default =
      examples_dir + "/media/pred_image.jpg";
  std::string input_file_path_default = examples_dir + "/media/bicycle.jpg";
  if (output_file_path.empty()) {
    output_file_path = output_file_path_default;
  }
  if (input_file_path.empty()) {
    input_file_path = input_file_path_default;
  }

  // Open image
  auto img = cv::imread(input_file_path, cv::IMREAD_COLOR);

  if (img.empty()) {
    std::cout << "Couldn't open image " << input_file_path
              << ". File or directory doesn't exist\n";
    return 1;
  }

  std::cout << "Test parameters: " << std::endl;
  std::cout << "Input file: " << input_file_path << std::endl;
  std::cout << "Output file: " << output_file_path << std::endl;
  std::cout << "Score threshold: " << score_threshold << std::endl;
  std::cout << "IOU NMS threshold: " << nms_threshold << std::endl;
  std::cout << "Confidence threshold: " << confidence_threshold << std::endl;

  auto backend_id = InitBackend(1);

  std::string model_json = examples_dir + "/models/yolov5s-model.json";
  std::string model_weights = examples_dir + "/models/yolov5s-weights.bin";
  auto model_id = LoadModel(model_json.c_str(), model_weights.c_str(),
                            ENNDataType::FLOAT16, backend_id);

  // Preparing data before calculations
  cv::Size origin_size;
  cv::Mat src = preprocess_image(img, origin_size);

  // Creating destination matrix with yolov5 output shape
  cv::Mat dst = cv::Mat::zeros(25200, 85, CV_32FC1);

  float* inputs[1] = {reinterpret_cast<float*>(src.data)};
  float* outputs[1] = {reinterpret_cast<float*>(dst.data)};

  // measure invoke time
  auto start_time = std::chrono::steady_clock::now();

  InvokeModel(model_id, inputs, outputs, 1);

  auto end_time = std::chrono::steady_clock::now();

  std::vector<Detection> preds = {};
  output_proc(dst, preds, coco_classes, score_threshold, nms_threshold,
              confidence_threshold);

  // scale coeff between original image and yolov5 640x640 input image
  float scale = (float)std::max(img.size().height, img.size().width) / 640;

  // Print predictions
  for (int i = 0; i < preds.size(); ++i)
    std::cout
        << "_____________\nClass: " << coco_classes[preds[i].class_id].first
        << "\n"
        << "Confidence: " << preds[i].confidence << "\n"
        << "Boundbox: " << preds[i].box << std::endl;

  // Draw predicted bounding-boxes on image
  for (auto& pred : preds) {
    int id = pred.class_id;
    std::stringstream converter;
    converter << ceil(pred.confidence * 1e2) / 1e2;
    cv::Size text_size =
        cv::getTextSize(coco_classes[id].first + ": " + converter.str(),
                        cv::FONT_HERSHEY_SIMPLEX, 0.7, 1, nullptr);
    int text_width = text_size.width;
    rectangle(img, pred.box.tl() * scale, pred.box.br() * scale,
              coco_classes[id].second, 2);
    rectangle(img, (pred.box.tl() + cv::Point(0, 20)) * scale,
              (pred.box.tl() + cv::Point(text_width / scale + 2, -1)) * scale,
              coco_classes[id].second, -1);
    putText(img, coco_classes[id].first + ": " + converter.str(),
            (pred.box.tl() + cv::Point(1, 15)) * scale,
            cv::FONT_HERSHEY_SIMPLEX, 0.7, cv::Scalar(255, 255, 255), 2);
  }

  std::cout << std::endl
            << "Prediction time: "
            << std::chrono::duration_cast<std::chrono::milliseconds>(
                   end_time - start_time)
                   .count()
            << " ms." << std::endl;

  if (cv::imwrite(output_file_path, img)) {
    std::cout << std::endl
              << "Image saved as " << output_file_path << std::endl;
  } else {
    std::cout << "Cannot save result into " << output_file_path
              << ". File or directory doesn't exits." << std::endl;
  }

  ReleaseModel(model_id);
  ReleaseBackend(backend_id);

  return 0;
}
